@@ -93,12 +93,39 @@ pub struct Connection
 {
     inner: Arc<Mutex<InnerConnection>>,
     sem: Arc<Semaphore>,
-    addr: FCGIAddr
+    addr: FCGIAddr,
+    header_mul: MultiHeaderStrategy,
+    header_nl: HeaderMultilineStrategy
+}
+#[derive(Copy, Clone)]
+pub enum MultiHeaderStrategy {
+    Combine,
+    OnlyFirst,
+    OnlyLast
+}
+#[derive(Copy, Clone)]
+pub enum HeaderMultilineStrategy {
+    Ignore,
+    ReturnError
 }
 impl Connection
 {
     /// Connect to a peer
-    pub async fn connect(addr: &FCGIAddr, max_req_per_con: u16) -> Result<Connection, Box<dyn Error>> {
+    #[inline]
+    pub async fn connect(addr: &FCGIAddr,
+        max_req_per_con: u16)
+        -> Result<Connection, Box<dyn Error>> {
+        Self::connect_with_strategy(addr,
+            max_req_per_con,
+            MultiHeaderStrategy::OnlyFirst,
+            HeaderMultilineStrategy::Ignore).await
+    }
+    /// Connect to a peer
+    pub async fn connect_with_strategy(addr: &FCGIAddr,
+        max_req_per_con: u16,
+        header_mul: MultiHeaderStrategy,
+        header_nl: HeaderMultilineStrategy)
+         -> Result<Connection, Box<dyn Error>> {
         Ok(Connection {
             inner: Arc::new(Mutex::new(InnerConnection{
                 io: FCGIWriter::new(BufReader::new(Stream::connect(addr).await?)),
@@ -106,7 +133,9 @@ impl Connection
                 fcgi_parser: fastcgi::RecordReader::new()
             })),
             sem: Arc::new(Semaphore::new(max_req_per_con as  usize)),
-            addr: addr.clone()
+            addr: addr.clone(),
+            header_mul,
+            header_nl
         })
     }
 
@@ -123,6 +152,11 @@ impl Connection
         Ok(())
     }
 
+    const QUERY_STRING: &'static [u8] = b"QUERY_STRING";
+    const REQUEST_METHOD: &'static [u8] = b"REQUEST_METHOD";
+    const CONTENT_TYPE: &'static [u8] = b"CONTENT_TYPE";
+    const CONTENT_LENGTH: &'static [u8] = b"CONTENT_LENGTH";
+    const NULL: &'static [u8] = b"";
     /// Forwards an HTTP request to a FGCI Application
     ///
     /// Fills QUERY_STRING, REQUEST_METHOD, CONTENT_TYPE and CONTENT_LENGTH
@@ -149,12 +183,14 @@ impl Connection
     /// - REQUEST_URI       common
     /// - DOCUMENT_URI      common
     /// - DOCUMENT_ROOT     common
-    pub async fn forward<B, I>(&self,
+    pub async fn forward<B, I, P1, P2>(&self,
                         req: Request<B>,
                         dyn_headers: I)
                     -> Result<Response<impl Body<Data = Bytes,Error = IoError>>, IoError>
     where   B: Body+Unpin,
-            I: IntoIterator<Item = (Bytes, Bytes)>
+            I: IntoIterator<Item = (P1,P2)>,
+            P1: Buf,
+            P2: Buf
     {
         let rid: u16;
         {
@@ -198,43 +234,108 @@ impl Connection
     
             kvw.extend(dyn_headers).await?;
 
-            let query = match req.uri().query() {
-                Some(query) => BytesMut::from(query.as_bytes()).freeze(),
-                None => Bytes::new()
-            };
-            kvw.add(fastcgi::NameValuePair::new(Bytes::from(&b"QUERY_STRING"[..]),query)).await?; //must CGI1.1 4.1.7
+            match req.uri().query(){
+                Some(query) => kvw.add_kv(Self::QUERY_STRING, query.as_bytes()).await?, //must CGI1.1 4.1.7
+                None => kvw.add_kv(Self::QUERY_STRING, Self::NULL).await? //must CGI1.1 4.1.7
+            }
         
-            let method = BytesMut::from(req.method().as_str().as_bytes()).freeze();
-            kvw.add(fastcgi::NameValuePair::new(Bytes::from(&b"REQUEST_METHOD"[..]),method)).await?; //must CGI1.1 4.1.12
+            kvw.add_kv(Self::REQUEST_METHOD, req.method().as_str().as_bytes()).await?; //must CGI1.1 4.1.12
             
-            if let Some(value) = req.headers().get(CONTENT_TYPE) { //if client CGI1.1 4.1.3.
-                kvw.add(fastcgi::NameValuePair::new(
-                        Bytes::from(&b"CONTENT_TYPE"[..]),
-                        BytesMut::from(value.as_bytes()).freeze()
-                    )).await?;
-            }
-            let len = req.headers().get(CONTENT_LENGTH); //if body CGI1.1 4.1.2.
-            if let Some(value) = len {
-                kvw.add(fastcgi::NameValuePair::new(
-                        Bytes::from(&b"CONTENT_LENGTH"[..]),
-                        BytesMut::from(value.as_bytes()).freeze()
-                    )).await?;
-            }
-            let skip = [AUTHORIZATION, CONTENT_LENGTH, CONTENT_TYPE];
-            //append all HTTP headers
-            for (key, value) in req.headers().iter() {
-                if skip.iter().find(|x| x == key).is_some() { //CGI1.1 4.1.18.
-                    continue;
+            let (len, body) = {
+                let (parts, body) = req.into_parts();
+                let headers = parts.headers;
+
+                if let Some(value) = headers.get(CONTENT_TYPE) { //if client CGI1.1 4.1.3.
+                    kvw.add_kv(
+                            Self::CONTENT_TYPE,
+                            value.as_bytes()
+                        ).await?;
                 }
-                let mut k = BytesMut::with_capacity(key.as_str().len()+5);
-                k.put(&b"HTTP_"[..]);
-                k.put(key.as_str().as_bytes()); //FIXME copy
-                let v = BytesMut::from(value.as_bytes()).freeze();
-                let p = fastcgi::NameValuePair::new(k.freeze(),v);
-                kvw.add(p).await?;
-            }
-            //send all headers to the FCGI App
-            kvw.flush().await?;
+                
+                let len: Option<usize> = if let Some(value) = headers.get(CONTENT_LENGTH) { //if body CGI1.1 4.1.2.
+                    kvw.add_kv(
+                            Self::CONTENT_LENGTH,
+                            value.as_bytes()
+                        ).await?;
+                    Some(value.to_str().ok()
+                        .and_then(|s|s.parse().ok())
+                        .unwrap_or(0))
+                }else{
+                    None
+                };
+                let skip = [AUTHORIZATION, CONTENT_LENGTH, CONTENT_TYPE];
+                //append all HTTP headers
+                for key in headers.keys() {
+                    if skip.iter().find(|x| x == key).is_some() { //CGI1.1 4.1.18.
+                        continue;
+                    }
+                    /*rfc3875
+                    The HTTP header field name is converted to upper case, has all
+                    occurrences of "-" replaced with "_" and has "HTTP_" prepended to
+                    give the meta-variable name.
+                    */
+                    let mut k = BytesMut::with_capacity(key.as_str().len()+5);
+                    k.put(&b"HTTP_"[..]);
+                    for &c in key.as_str().as_bytes() {
+                        let upper = match c {
+                            b'-' => b'_',
+                            lower_acii if b'a' <= lower_acii && lower_acii <= b'z' => lower_acii - (b'a'-b'A'), //a ... z
+                            s => s
+                        };
+                        k.put_u8(upper);
+                    }
+                    /*rfc3875
+                    If multiple header fields with the same field-name
+                    are received then the server MUST rewrite them as a single value
+                    having the same semantics.  Similarly, a header field that spans
+                    multiple lines MUST be merged onto a single line.
+
+                    RFC 7230, Section 3.2.2, Field Order: Set-Cookie is special
+                    -> but not part of a request
+
+                    RFC 7230, Section 3.2.4, Field Parsing: multiline header -> 400/502
+                    */
+                    let mut value_buf;
+                    let value = match self.header_mul {
+                        MultiHeaderStrategy::Combine => {
+                            value_buf = BytesMut::with_capacity(512);
+                            let mut first = false;
+                            for v in headers.get_all(key).iter() {
+                                if !first {
+                                    first = true;
+                                }else{
+                                    value_buf.put_u8(b',');
+                                }
+                                let v = v.as_bytes();
+                                value_buf.put_slice(v);//copy
+                            }
+                            value_buf.as_ref()
+                        },
+                        MultiHeaderStrategy::OnlyFirst => {
+                            match headers.get(key) {
+                                Some(v) => v.as_bytes(),
+                                None => Self::NULL
+                            }
+                        },
+                        MultiHeaderStrategy::OnlyLast => {
+                            match headers.get_all(key).iter().next_back() {
+                                Some(v) => v.as_bytes(),
+                                None => Self::NULL
+                            }
+                            
+                        }
+                    };
+                    if let HeaderMultilineStrategy::ReturnError = self.header_nl {
+                        if value.as_ref().contains(&b'\n') {
+                            return Err(IoError::new(ErrorKind::InvalidData,"multiline headers are not allowed"));
+                        }
+                    }
+                    kvw.add_kv(k, value).await?;
+                }
+                //send all headers to the FCGI App
+                kvw.flush().await?;
+                (len, body)
+            };
             //mut_inner.io.encode(params).await?;
             trace!("sent header");
             //Note: Responses might arrive from this point on
@@ -242,14 +343,9 @@ impl Connection
             // FIXME stdin after end!!!!!!!
 
             //send the body to the FCGI App
-            if let Some(value) = len {
+            if let Some(len) = len {
                 //CGI1.1 4.2 -> at least content-length data
-                let len: usize = if let Ok(vstr) = value.to_str() {
-                    vstr.parse().unwrap_or(0)
-                }else{
-                    0
-                };
-                mut_inner.io.data_stream(req.into_body(), 
+                mut_inner.io.data_stream(body,
                     rid, 
                     fastcgi::Record::STDIN,
                     len).await?;
@@ -654,7 +750,7 @@ mod tests {
             let mut buf = BytesMut::with_capacity(4096);
             app_socket.read_buf(&mut buf).await.unwrap();
             trace!("app read {:?}", buf);
-            let to_php = b"\x01\x01\0\x01\0\x08\0\0\0\x01\x01\0\0\0\0\0\x01\x04\0\x01\0i\x07\0\x0f\x1cSCRIPT_FILENAME/home/daniel/Public/test.php\x0c\x05QUERY_STRINGlol=1\x0e\x03REQUEST_METHODGET\x0b\tHTTP_accepttext/html\x01\x04\0\x01\0i\x07\x01\x04\0\x01\0\0\0\0\x01\x05\0\x01\0\0\0\0";
+            let to_php = b"\x01\x01\0\x01\0\x08\0\0\0\x01\x01\0\0\0\0\0\x01\x04\0\x01\0i\x07\0\x0f\x1cSCRIPT_FILENAME/home/daniel/Public/test.php\x0c\x05QUERY_STRINGlol=1\x0e\x03REQUEST_METHODGET\x0b\tHTTP_ACCEPTtext/html\x01\x04\0\x01\0i\x07\x01\x04\0\x01\0\0\0\0\x01\x05\0\x01\0\0\0\0";
             assert_eq!(buf, Bytes::from(&to_php[..]));
             trace!("app answers on get");
             let from_php = b"\x01\x07\0\x01\0W\x01\0PHP Fatal error:  Kann nicht durch 0 teilen in /home/daniel/Public/test.php on line 14\n\0\x01\x06\0\x01\x01\xf7\x01\0Status: 404 Not Found\r\nX-Powered-By: PHP/7.3.16\r\nX-Authenticate: NTLM\r\nContent-type: text/html; charset=UTF-8\r\n\r\n<html><body>\npub\n<pre>Array\n(\n)\nArray\n(\n    [lol] => 1\n)\nArray\n(\n    [lol] => 1\n)\nArray\n(\n    [HTTP_accept] => text/html\n    [REQUEST_METHOD] => GET\n    [QUERY_STRING] => lol=1\n    [SCRIPT_NAME] => /test\n    [SCRIPT_FILENAME] => /home/daniel/Public/test.php\n    [FCGI_ROLE] => RESPONDER\n    [PHP_SELF] => /test\n    [REQUEST_TIME_FLOAT] => 1587740954.2741\n    [REQUEST_TIME] => 1587740954\n)\n\0\x01\x03\0\x01\0\x08\0\0\0\0\0\0\0\0\0\0";
@@ -680,7 +776,7 @@ mod tests {
             );
             let mut res = fcgi_con.forward(req,params).await.expect("forward failed");
             trace!("got res obj");
-            assert_eq!(res.status(), StatusCode::NOT_FOUND);//FIXME 200
+            assert_eq!(res.status(), StatusCode::NOT_FOUND);
             assert_eq!(res.headers().get("X-Powered-By").expect("powered by header missing"), "PHP/7.3.16");
             let read1 = res.data().await;
             assert!(read1.is_some());
@@ -722,7 +818,7 @@ mod tests {
             };
             let req = Request::get("/").body(b).unwrap();
             trace!("new req obj");
-            let params = HashMap::new();
+            let params: HashMap<Bytes,Bytes> = HashMap::new();
             let mut res = fcgi_con.forward(req,params).await.expect("forward failed");
             trace!("got res obj");
             let read1 = res.data().await;
@@ -766,7 +862,7 @@ mod tests {
             };
             let req = Request::get("/").body(b).unwrap();
             trace!("new req obj");
-            let params = HashMap::new();
+            let params: HashMap<Bytes,Bytes> = HashMap::new();
             let mut res = fcgi_con.forward(req,params).await.expect("forward failed");
             trace!("got res obj");
             assert_eq!(res.status(), StatusCode::OK);
@@ -793,7 +889,7 @@ mod tests {
             let to_php = b"\x01\x01\0\x01\0\x08\0\0\0\x01\x01\0\0\0\0\0\x01\x04\0\x01\0\x81\x07\0\x0f\x1cSCRIPT_FILENAME/home/daniel/Public/test.php\x0c\0QUERY_STRING\x0e\x04REQUEST_METHODPOST\x0c\x13CONTENT_TYPEmultipart/form-data\x0e\x01CONTENT_LENGTH8\x01\x04\0\x01\0\x81\x07\x01\x04\0\x01\0\0\0\0\x01\x05\0\x01\0\x08\0\0test=123\x01\x05\0\x01\0\0\0\0";
             assert_eq!(buf, Bytes::from(&to_php[..]));
             trace!("app answers on get");
-            let from_php = b"\x01\x07\0\x01\0W\x01\0PHP Fatal error:  Kann nicht durch 0 teilen in /home/daniel/Public/test.php on line 14\n\0\x01\x06\0\x01\x01\xf7\x01\0Status: 404 Not Found\r\nX-Powered-By: PHP/7.3.16\r\nX-Authenticate: NTLM\r\nContent-type: text/html; charset=UTF-8\r\n\r\n<html><body>\npub\n<pre>Array\n(\n)\nArray\n(\n    [lol] => 1\n)\nArray\n(\n    [lol] => 1\n)\nArray\n(\n    [HTTP_accept] => text/html\n    [REQUEST_METHOD] => GET\n    [QUERY_STRING] => lol=1\n    [SCRIPT_NAME] => /test\n    [SCRIPT_FILENAME] => /home/daniel/Public/test.php\n    [FCGI_ROLE] => RESPONDER\n    [PHP_SELF] => /test\n    [REQUEST_TIME_FLOAT] => 1587740954.2741\n    [REQUEST_TIME] => 1587740954\n)\n\0\x01\x03\0\x01\0\x08\0\0\0\0\0\0\0\0\0\0";
+            let from_php = b"\x01\x06\0\x01\x00\x23\x05\0Status: 201 Created\r\n\r\n<html><body>#+#+#\x01\x03\0\x01\0\x08\0\0\0\0\0\0\0\0\0\0";
             app_socket.write_buf(&mut Bytes::from(&from_php[..])).await.unwrap();
         }
         
@@ -817,14 +913,13 @@ mod tests {
             );
             let mut res = fcgi_con.forward(req,params).await.expect("forward failed");
             trace!("got res obj");
-            assert_eq!(res.status(), StatusCode::NOT_FOUND);//FIXME 200
-            assert_eq!(res.headers().get("X-Powered-By").expect("powered by header missing"), "PHP/7.3.16");
+            assert_eq!(res.status(), StatusCode::CREATED);
             let read1 = res.data().await;
             assert!(read1.is_some());
             let read1 = read1.unwrap();
             assert!(read1.is_ok());
             if let Ok(d) = read1 {
-                let body = b"<html><body>\npub\n<pre>Array\n(\n)\nArray\n(\n    [lol] => 1\n)\nArray\n(\n    [lol] => 1\n)\nArray\n(\n    [HTTP_accept] => text/html\n    [REQUEST_METHOD] => GET\n    [QUERY_STRING] => lol=1\n    [SCRIPT_NAME] => /test\n    [SCRIPT_FILENAME] => /home/daniel/Public/test.php\n    [FCGI_ROLE] => RESPONDER\n    [PHP_SELF] => /test\n    [REQUEST_TIME_FLOAT] => 1587740954.2741\n    [REQUEST_TIME] => 1587740954\n)\n";
+                let body = b"<html><body>";
                 assert_eq!(d, &body[..] );
             }
             let read2 = res.data().await;
